import java.io.*;
import java.util.*;
import javax.microedition.io.*;
import javax.microedition.lcdui.*;

public class VidanceCanvas extends Canvas implements Runnable, CommandListener {
    private String error;
    private StreamConnection sc;
    private DataInputStream dis;
    private DataOutputStream dos;
    private Image img;
    private Font font;
    private byte[] tmp, msg;
    private int colormode;
    private int xp, yp, width, height, swidth, sheight;
    private int bpp, rmax, gmax, bmax, mr,mg,mb,nr,ng,nb,dr,dg,db;
    private int rshift, gshift, bshift, truecolor;
    private int bigend, depth;
    private int pxsize, dmode, devdepth, devcolors;
    private int[] palette;
    private boolean fixpalette=false;
    private boolean running;
    private int menusize, ms34;
    private int mousebuttons=0, lastx, lasty;
    private int status=0;
    private String desktop;
    private boolean scrolled;
    private Display disp;
    
    private TextBox tb;
    private Command tbok=new Command("Ok",Command.OK,1), tbnm=new Command("Cancel",Command.CANCEL,2), inpt=new Command("Input",Command.OK,1), dcon=new Command("Disconnect",Command.BACK,2);
    
    public VidanceCanvas() {
        addCommand(inpt);
        addCommand(dcon);
        setCommandListener(this);
    }
    public void init() {
        width=getWidth();
        height=getHeight();
        menusize=getHeight()/8;
        menusize=Math.max(8,menusize);
        menusize=Math.min(20,menusize);
        height-=menusize;
        ms34=menusize*3/4;
        xp=0; yp=0;
        img=Image.createImage(width,height);
        tmp=new byte[128];
        msg=new byte[24];
        colormode=0;
        font=Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, Font.SIZE_SMALL);
        error=null;
        running=true;
        scrolled=true;
        try { sc.close(); } catch(Exception e) {}
        if(tb==null) {
            tb=new TextBox("Vidance Input","",120,TextField.ANY);
            tb.addCommand(tbok);
            tb.addCommand(tbnm);
            tb.setCommandListener(this);
        }
    }
    /** Set color mode. Must be called before connect. */
    public void setColorMode(int n) {
        colormode=n;
    }
    /** Set display. Used to determine color modes. */
    public void setDisplay(Display d) {
        disp=d;
    }
    public void paint(Graphics g) {
        g.setColor(0xFFFFFF);
        g.fillRect(0,height,width,menusize);

        if(status==1) {
            g.setColor(0);
            g.fillRect(width-ms34+2,height+2,ms34-2,menusize-2);
        } else {
            g.drawImage(img,0,0,g.LEFT|g.TOP);
            g.setColor(0);
            //draw mouse button indicators
            if((mousebuttons&1)!=0) 
                g.fillArc(0,height,ms34*2,menusize*2,90,90);
            else
                g.drawArc(0,height,ms34*2,menusize*2,90,90);
            if((mousebuttons&2)!=0) 
                g.fillRect(ms34,height,ms34,menusize);
            else 
                g.drawRect(ms34,height,ms34,menusize);
            if((mousebuttons&4)!=0) 
                g.fillArc(ms34,height,ms34*2,menusize*2,0,90);
            else 
                g.drawArc(ms34,height,ms34*2,menusize*2,0,90);

            g.setFont(font);
            g.drawString("E",ms34*3,height,g.TOP|g.LEFT);
            g.drawString("B",ms34*4,height,g.TOP|g.LEFT);
        }
    }
    public int connect(String host, String pw) {
        int major=0, minor=0, t=0;
        try {
            sc=(StreamConnection)Connector.open("socket://"+host);
            dis=sc.openDataInputStream();
            dos=sc.openDataOutputStream();
            dis.readFully(tmp,0,12);
            for(int i=4; i<7; i++) {
                major=major*10+(tmp[i]-0x30);
                minor=minor*10+(tmp[i+4]-0x30);
            }
            System.err.println("Server "+major+"."+minor);
            dos.write(new String("RFB 003.003\n").getBytes());
            t=dis.readInt();
            if(t==0) return 1; //connection failed
            if(t==2) {
                System.err.println("VNC auth");
                dis.readFully(tmp,0,16);
                VNCEncrypt.vncMagic(tmp,pw);
                System.gc();
                dos.write(tmp,0,16);
                t=dis.readInt();
                if(t>0) {
                    System.err.println("Auth failed: "+t);
                    return 2; //auth failed
                }
            }
            dos.write(1); //share connection

            swidth=dis.readUnsignedShort();
            sheight=dis.readUnsignedShort();
            bpp=dis.read();
            depth=dis.read();
            bigend=dis.read();
            truecolor=dis.read();
            rmax=dis.readUnsignedShort();
            gmax=dis.readUnsignedShort();
            bmax=dis.readUnsignedShort();
            rshift=dis.read();
            gshift=dis.read();
            bshift=dis.read();
            
            System.err.println(bpp+" bpp");
            System.err.println(depth+" depth");
            System.err.println(bigend+" bigendian");
            System.err.println(truecolor+" truecolor");
            System.err.println(rmax+" redmax");
            System.err.println(gmax+" greenmax");
            System.err.println(bmax+" bluemax");
            System.err.println(rshift+" redshift");
            System.err.println(gshift+" greenshift");
            System.err.println(bshift+" blueshift");
            
            t=101; //for error handler
            
            switch(colormode) {
                case 0: //keep servers.
                    break;
                case 1: //best guess. this is all pulled out of my ass.
                    System.err.println("Best guess");
                    int d=0,n=1; 
                    for(int i=1; i<32; i++) {
                        n=n*2;
                        if(n==disp.numColors()) {
                            d=i;
                            break;
                        }
                    }
                    devdepth=d;
                    devcolors=disp.numColors();
                    if(d==0) { //weird color count, we'll just bgr233
                        bpp=8; depth=8; bigend=1; truecolor=1;
                        rmax=7; gmax=7; bmax=3;
                        rshift=0; gshift=3; bshift=6;
                        break;
                    }
                    if(disp.isColor()) {
                        if(d>18) { //more than 18 bits, use 24
                            bpp=32; depth=24; bigend=1; truecolor=1;
                            rmax=255; gmax=255; bmax=255;
                            rshift=16; gshift=8; bshift=0;
                        } else if(d>12) { //12-18, use 16 bit
                            bpp=16; depth=16; bigend=1; truecolor=1;
                            rmax=63; gmax=31; bmax=31;
                            rshift=10; gshift=5; bshift=0;
                        } else if(d>8) { //8-12, use 12bit map
                            bpp=8; depth=12; bigend=1; truecolor=0;
                        } else { //else bgr233
                            bpp=8; depth=8; bigend=1; truecolor=1;
                            rmax=7; gmax=7; bmax=3;
                            rshift=0; gshift=3; bshift=6;
                        }
                    } else { //grayscale
                        if(d>8) { //use 15bit
                            bpp=16; depth=15; bigend=1; truecolor=1;
                            rmax=31; gmax=31; bmax=31;
                            rshift=10; gshift=5; bshift=0;
                        } else {
                            fixpalette=true;
                            bpp=8; depth=d; bigend=1; truecolor=0;
                        }
                    }
                    break;
                case 2: //bgr233
                    System.err.println("BGR233");
                    bpp=8; depth=8; bigend=1; truecolor=1;
                    rmax=7; gmax=7; bmax=3; 
                    rshift=0; gshift=3; bshift=6;
                    break;
                case 3: //8bit map (48bit)
                    System.err.println("8bit map");
                    bpp=8; depth=48; bigend=1; truecolor=0;
//                    rmax=65535; gmax=65535; bmax=65535; 
//                    rshift=0; gshift=16; bshift=32;
                    break;
                case 4: //16bit
                    System.err.println("16bit");
                    bpp=16; depth=16; bigend=1; truecolor=1;
                    rmax=63; gmax=31; bmax=31; 
                    rshift=10; gshift=5; bshift=0;
                    break;
                case 5: //24bit depth, 32bit color
                    System.err.println("24bit");
                    bpp=32; depth=24; bigend=1; truecolor=1;
                    rmax=255; gmax=255; bmax=255; 
                    rshift=16; gshift=8; bshift=0;
                    break;
            }
            System.err.println(".............");
            System.err.println(bpp+" bpp");
            System.err.println(depth+" depth");
            System.err.println(bigend+" bigendian");
            System.err.println(truecolor+" truecolor");
            System.err.println(rmax+" redmax");
            System.err.println(gmax+" greenmax");
            System.err.println(bmax+" bluemax");
            System.err.println(rshift+" redshift");
            System.err.println(gshift+" greenshift");
            System.err.println(bshift+" blueshift");
            
            for(int i=0; i<3; i++) dis.read();
            t=dis.readInt();
            dis.readFully(tmp,0,t);
            desktop=new String(tmp,0,t);
            
            if(colormode!=0) {
                synchronized(dos) {
                    dos.write(0);
                    for(int i=0; i<3; i++) dos.write(0);
                    dos.write(bpp);
                    dos.write(depth);
                    dos.write(bigend);
                    dos.write(truecolor);
                    dos.writeShort(rmax);
                    dos.writeShort(gmax);
                    dos.writeShort(bmax);
                    dos.write(rshift);
                    dos.write(gshift);
                    dos.write(bshift);
                    for(int i=0; i<3; i++) dos.write(0);
                }
            }
            
            synchronized(dos) {
                dos.write(2); //accepted encodings
                dos.write(0);
                dos.writeShort(2);
                dos.writeInt(4);//CoRRE
                dos.writeInt(0);//raw
            }
            
            t=calcColors();
            if(t==-1) return 3;
            
            new Thread(this).start();
            return 0;
        } catch(Exception e) {
            error=e.toString();
            if(t==101) error+=", tried another color mode?";
            e.printStackTrace();
            return -1;
        }
    }
    private int getColor(int px) {
/*
        int r,g,b;
        int or,og,ob;
        r=(px>>>rshift)&rmax;
        g=(px>>>gshift)&gmax;
        b=(px>>>bshift)&bmax;
        System.err.println("Real RGB:" +r+","+g+","+b);
        or=r*dr;
        og=g*dg;
        ob=b*db;
        System.err.println("255 RGB:" +or+","+og+","+ob);
        return (or<<16) | (og<<8) | ob;
*/        
        return (((px>>>rshift)&rmax)<<nr) | (((px>>>gshift)&gmax)<<ng) | (((px>>>bshift)&bmax)<<nb);
    }
    private int calcColors() {
        int r,g,b,t;
        pxsize=bpp>>3; //assume no broken bytes
        palette=null;
        for(t=0; t<32; t++) 
            if(((1<<t)&rmax) == 0) break;
        mr=t;
        for(t=0; t<32; t++) 
            if(((1<<t)&gmax) == 0) break;
        mg=t;
        for(t=0; t<32; t++) 
            if(((1<<t)&bmax) == 0) break;
        mb=t;
        
        nr=24-mr;
        ng=16-mg;
        nb=8-mb;
        dr=0xFF/rmax;
        dg=0xFF/gmax;
        db=0xFF/bmax;
        
        if(pxsize>4) {
            System.err.println("Too many colors: "+bpp+"b, "+pxsize);
            return -1;
        } if(pxsize==1) {
            palette=new int[256];
            if(truecolor==1) {
                for(int i=0; i<256; i++) {
                    r=(i>>>rshift)&rmax;
                    g=(i>>>gshift)&gmax;
                    b=(i>>>bshift)&bmax;
                    palette[i]=(r*dr<<16) | (g*dg<<8) | (b*db);
                }
            } else if(fixpalette && devcolors<256) {
                t=256/devcolors;
                for(int i=0; i<devcolors; i++) {
                    r=t*i;
                    palette[i]=(r<<16) | (r<<8) | r;
                }
//                fixPalette(0,devcolors); //Unsupported by servers
            }
            dmode=1; //palette
        } else {
            System.err.println("Crazy color scheme");
            dmode=0; //manual
        }
        System.gc();
        System.err.println("Pixsize: "+pxsize+", Drawmode: "+dmode);
        return 0;
    }
    public void fixPalette(int off, int len) {
        System.err.println("Fixpalette: "+off+", "+len);
        int r,g,b;
        synchronized(dos) {
            try {
                dos.write(1);
                dos.write(0);
                dos.writeShort(off);
                dos.writeShort(len);
                for(int i=0; i<len; i++) {
                    r=(palette[off+i]>>>8)&0xFF00;
                    g=(palette[off+i])&0xFF00;
                    b=(palette[off+i]<<8)&0xFF00;
                    dos.writeShort(r);
                    dos.writeShort(g);
                    dos.writeShort(b);
                }
            } catch(IOException e) { e.printStackTrace(); }
        }
    }
    public void update() throws IOException {
        synchronized(dos) {
            dos.write(3);
            dos.write(scrolled?0:1);
            dos.writeShort(xp);
            dos.writeShort(yp);
            dos.writeShort(width);
            dos.writeShort(height);
            scrolled=false;
        }
    }
    public void sendPointer(int x, int y, int bm) throws IOException {
        synchronized(dos) {
            lastx=x;
            lasty=y;

            //Speed up mouse pointer by doing some buffering. 
            msg[0]=5;
            msg[1]=(byte)bm;
            msg[2]=(byte)(x>>>8);
            msg[3]=(byte)x;
            msg[4]=(byte)(y>>>8);
            msg[5]=(byte)y;
            dos.write(msg,0,6);

            /* 
            dos.write(5);
            dos.write(bm);
            dos.writeShort(x+xp);
            dos.writeShort(y+yp);
            */
        }
    }
    
    public void scroll(int dir) {
        synchronized(dos) {
            scrolled=true;
            switch(dir) {
                case 0: //up
                    yp-=height*4/5;
                    yp=Math.max(0,yp);
                    break;
                case 1: //down
                    yp+=height*4/5;
                    yp=Math.min(yp,sheight-height);
                    break;
                case 2: //left
                    xp-=width*4/5;
                    xp=Math.max(0,xp);
                    break;
                case 3: //right
                    xp+=width*4/5;
                    xp=Math.min(xp,swidth-width);
                    break;
            }
        }
    }
    protected void keyPressed(int keycode) {
        int n=getGameAction(keycode);
        switch(n) {
            case UP:
                scroll(0);
                break;
            case DOWN:
                scroll(1);
                break;
            case LEFT:
                scroll(2);
                break;
            case RIGHT:
                scroll(3);
                break;
            case FIRE:
               try { 
                   update(); 
               } catch(Exception e) { 
                   running=false;
                   error=e.toString();
                   e.printStackTrace();
                   notifyAll();
               }
               break;
        }
    }
    protected void pointerDragged(int x, int y) {
        try { 
            if(y<height) sendPointer(x+xp,y+yp,mousebuttons); 
        } catch(Exception e) {
            running=false;
            error=e.toString();
            e.printStackTrace();
            notifyAll();
        }
    }
    protected void pointerPressed(int x, int y) {
        if(y>height) {
            switch(x/ms34) {
                case 0:
                    mousebuttons^=1;
                    break;
                case 1:
                    mousebuttons^=2;
                    break;
                case 2:
                    mousebuttons^=4;
                    break;
                case 3: //enter
                    synchronized(dos) {
                        try {
                            setupChars();
                            sendChar(0xFF0D);
                        } catch(Exception e) {}
                    }
                    break;
                case 4: //backspace
                    synchronized(dos) {
                        try {
                            setupChars();
                            sendChar(0xFF08);
                        } catch(Exception e) {}
                    }
                    break;
            }
            repaint();
            serviceRepaints();
        } else {
            try { 
                sendPointer(x+xp,y+yp,mousebuttons); 
            } catch(Exception e) {
                running=false;
                error=e.toString();
                e.printStackTrace();
                notifyAll();
            }
        }
    }
    protected void pointerReleased(int x, int y) {
        try { 
            if(y<height) sendPointer(x+xp,y+yp,0); 
        } catch(Exception e) {
            running=false;
            error=e.toString();
            e.printStackTrace();
            notifyAll();
        }
    }
    public String getError() { return error; }
    public boolean hasError() { return error!=null; }

    public void run() {
        int packtype, rectnum, xpos, ypos, w, h, ow, oh, enc;
        int px, fpx;
        int t, u, v;
        Graphics g;
        g=img.getGraphics();
        g.setColor(0x000000);
        g.fillRect(10,10,20,20);
        g.setFont(font);
        g.drawString("Connected to",0,34,g.TOP | g.LEFT);
        g.drawString(desktop,0,48,g.TOP | g.LEFT);
        g.drawString("Press fire.",0,62,g.TOP | g.LEFT);
        repaint();
        serviceRepaints();
        try {
            while(running) {
            packtype=dis.read();
            if(packtype==0) { //FramebufferUpdate
                status=1;
                repaint();
                dis.read();  //padding
                rectnum=dis.readUnsignedShort();
                for(int k=0; k<rectnum; k++) {
                    xpos=dis.readUnsignedShort()-xp;
                    ypos=dis.readUnsignedShort()-yp;
                    w=dis.readUnsignedShort();
                    h=dis.readUnsignedShort();
                    enc=dis.readInt();
//                    System.err.println(xpos+","+ypos+" "+w+"x"+h+": "+enc+" ("+dmode+")");
                    if(enc==0) {
//                      System.err.println("Raw: "+xpos+","+ypos+": "+w+"x"+h);
                        ow=w+xpos;
                        oh=h+ypos;
                        switch(dmode) {
                            case 0: //manual
                                for(int i=ypos; i<oh; i++)
                                    for(int j=xpos; j<ow; j++) {
                                        px=0;
                                        dis.readFully(tmp,0,pxsize);
                                        for(int l=0; l<pxsize; l++)
                                            px=(px<<8) | (tmp[l]&0xFF);
                                        px=getColor(px);
                                        g.setColor(px);
                                        g.drawRect(j,i,1,1);
                                    }
                                break;
                            case 1: //use palette
                                fpx=h*w;
                                px=0;
                                u=xpos;
                                v=ypos;
                                while(px<fpx) {
                                    t=dis.read(tmp,0,Math.min(tmp.length,fpx-px));
                                    for(int i=0; i<t; i++) {
                                        g.setColor(palette[tmp[i]&0xFF]);
                                        g.drawRect(u,v,1,1);
                                        px++;
                                        u++;
                                        if(u==ow) {
                                            u=xpos;
                                            v++;
                                        }
                                    }
                                }
                                break;
                        } //switch
                    } else if(enc==4) {
                        t=dis.readInt();
//                  System.err.println(t+" tiles, "+w+"x"+h+" "+xpos+","+ypos);
                        px=0;
                        switch(dmode) {
                            case 0:
                                for(int i=0; i<pxsize; i++)
                                    px=(px<<8)|dis.read();
                                px=getColor(px);
                                break;
                            case 1:
                                px=palette[dis.read()];
                                break;
                        }
                        g.setColor(px);
                        g.fillRect(xpos,ypos,w,h);
                        switch(dmode) {
                            case 0:
                                for(int j=0; j<t; j++) {
                                    px=0;
                                    dis.readFully(tmp,0,pxsize);
                                    for(int i=0; i<pxsize; i++) 
                                        px=(px<<8)|(tmp[i]&0xFF);
                                    g.setColor(getColor(px));
                                    u=dis.read()+xpos;
                                    v=dis.read()+ypos;
                                    g.fillRect(u,v,dis.read(),dis.read());
                                }
                                break;
                            case 1:
                                for(int j=0; j<t; j++) {
                                    dis.readFully(tmp,0,5);
                                    g.setColor(palette[tmp[0]&0xFF]);
                                    g.fillRect((tmp[1]&0xFF)+xpos,(tmp[2]&0xFF)+ypos,tmp[3]&0xFF,tmp[4]&0xFF);
//                                    System.err.println(((tmp[1]&0xFF)+xpos)+","+((tmp[2]&0xFF)+ypos)+": "+(tmp[3]&0xFF)+"x"+(tmp[4]&0xFF));
                                }
                                break;
                        }
                    } //enc
                    else System.err.println("Unknown encoding: "+enc);
                } //rectnum
                status=0;
                repaint();
                serviceRepaints();
            } else if(packtype==1) { //setcolourmapentries
                dis.read();
                t=dis.readUnsignedShort();
                u=dis.readUnsignedShort();
//                System.err.println("SetColourMapEntries: "+t+","+u);
                v=u+t;
                for(int i=t; i<v; i++) {
                    u=(dis.readUnsignedShort()/257);  //more accurate but slow
                    u=(dis.readUnsignedShort()/257) | (u<<8);
                    u=(dis.readUnsignedShort()/257) | (u<<8);
                    palette[i]=u;
                }
            } else if(packtype==2) {
                System.err.println("Bell!");
                //beep? visual alert?
            } else if(packtype==3) {
                System.err.println("Server cut text."); 
                for(int i=0; i<3; i++) dis.read();
                t=dis.readInt();
                dis.read(tmp,0,t);
                //keep?
            } else if(packtype==-1 && running) {
                error="End of stream.";
                running=false;
            } else System.out.println("Unknown packet"+packtype);
            } //running loop
        } catch(Exception e) {
            running=false;
            error=e.toString()+" in run()";
            e.printStackTrace();
        } finally {
            notifyAll();
        }
    }
    public void sendString(String s) throws IOException {
        synchronized(dos) {
            char r;
            setupChars();
            for(int i=0; i<s.length(); i++) {
                r=s.charAt(i);
                switch(r) {
                    case '\n':
                    case '\r':
                        sendChar(0xFF0D);
                        break;
                    default:
                        sendChar(r);
                        break;
                }
            }
        }
    }
    private void setupChars() {
        for(int i=0; i<16; i++)
            msg[i]=0;
        msg[0]=msg[8]=4;
        msg[1]=1; msg[9]=0;
    }
    /** Send a char. Do not call, use sendString instead. */
    private void sendChar(int c) throws IOException {
        msg[6]=msg[14]=(byte)((c>>>8)&0xFF);
        msg[7]=msg[15]=(byte)(c&0xFF);
        try { dos.write(msg,0,16); 
            System.err.println("Cow!");
        } catch(Exception e) {
            e.printStackTrace();
        }
        
    }
    public void commandAction(Command c, Displayable d) {
        System.gc();
        if(d==tb) {
            if(c==tbok) {
                try { sendString(tb.getString()); } catch(Exception e) {}
            } 
            disp.setCurrent(this);
        } else if(d==this) {
            if(c==inpt) displayTextInput();
            else if(c==dcon) {
                running=false;
                try { 
                    dos.write(0xFF); //let's try to get evicted. FIXME
                    sc.close(); 
                } catch(Exception e) {}
            }
        }
    }
    public void displayTextInput() {
        tb.setString("");
        disp.setCurrent(tb);
    }
}

